Option Explicit On
Option Strict On

Imports System.Math

Module MathInGammaF
    Private Const ITMAX As Integer = 100
    Private Const EPS As Double = 0.00000000000000022204460492503131
    Private Const FPMIN As Double = 2.2250738585072014E-308 / EPS

    Public Function GammP(ByVal A As Double, ByVal X As Double) As Double
        If X < 0.0 Then Throw New System.ArgumentOutOfRangeException _
                                    ("X", X, "The argument must be 0.0 or positive.")
        If A <= 0.0 Then Throw New System.ArgumentOutOfRangeException _
                                    ("A", A, "The argument must be positive.")

        If X < A + 1.0 Then
            Return GSer(A, X)
        Else
            Return 1.0 - Gcf(A, X)
        End If
    End Function
    Public Function GammQ(ByVal A As Double, ByVal X As Double) As Double
        If X < 0.0 Then Throw New System.ArgumentOutOfRangeException _
                                    ("X", X, "The argument must be 0.0 or positive.")
        If A <= 0.0 Then Throw New System.ArgumentOutOfRangeException _
                                    ("A", A, "The argument must be positive.")

        If X < A + 1.0 Then
            Return 1.0 - GSer(A, X)
        Else
            Return Gcf(A, X)
        End If
    End Function
    Private Function GSer(ByVal A As Double, ByVal X As Double) As Double
        Dim Gln As Double = GammLn(A)

        If X <= 0.0 Then
            If X < 0.0 Then Throw New System.ArgumentOutOfRangeException _
                                        ("X", X, "The argument must be 0.0 or positive.")
            Return 0.0
        Else
            Dim ap As Double = A
            Dim del As Double = 1.0 / A
            Dim sum As Double = del

            For n As Integer = 1 To ITMAX
                ap += 1.0
                del *= X / ap
                sum += del
                If Abs(del) < Abs(sum) * EPS Then Return sum * Exp(-X + A * Log(X) - Gln)
            Next

            Throw New System.ArithmeticException("Bad A and ITMAX in GSer")
        End If
    End Function
    Private Function Gcf(ByVal A As Double, ByVal X As Double) As Double
        Dim Gln As Double = GammLn(A)

        Dim b As Double = X + 1.0 - A
        Dim c As Double = 1.0 / FPMIN
        Dim d As Double = 1.0 / b
        Dim h As Double = d

        For i As Integer = 1 To ITMAX
            Dim an As Double = -i * (i - A)
            b += 2.0
            d = an * d + b
            If Abs(d) < FPMIN Then d = FPMIN
            c = b + an / c
            If Abs(c) < FPMIN Then c = FPMIN
            d = 1.0 / d
            Dim del As Double = d * c
            h *= del
            If Abs(del - 1.0) <= EPS Then Return Exp(-X + A * Log(X) - Gln) * h
        Next
        Throw New System.ArithmeticException("Bad A and ITMAX in Gcf")
    End Function
End Module
